function [kh2,khi,gh2,ghi]=h2hi(w,u,e,y)
% function [kh2,khi,gh2,ghi]=h2hi(w,u,e,y)
%
% optimal CT H2 and H-Infinity design for 
% LTI system defined using abstract signals
% w,u,e,y: disturbance,control,cost and sensor signals
% tol: tolerance relative to h2 optimum, default tol=0.01

% written by ameg@mit.edu,  last modified March 10, 2005

if nargin<4, error('first 4 arguments required'); end
tol=0.0001;    



% check arguments
if ~(isa(w,'asgn')&isa(u,'asgn')&isa(e,'asgn')&isa(y,'asgn')), 
	error('first four arguments must be abstract signals')
end

global abstract_A;
global abstract_B;
global abstract_C;
global abstract_D;
global abstract_n;
global abstract_m;
global abstract_k;

rw=ref(w);
nw=size(w);
ru=ref(u);
nu=size(u);
re=ref(e);
ne=size(e);
ry=ref(y);
ny=size(y);

% input c coefficients
cci=[abstract_C(rw+1:rw+nw,:);abstract_C(ru+1:ru+nu,:)]; 
% input d coefficients
ddi=[abstract_D(rw+1:rw+nw,:);abstract_D(ru+1:ru+nu,:)];  
% output c
cco=[abstract_C(re+1:re+ne,:);abstract_C(ry+1:ry+ny,:)];  
% output d
ddo=[abstract_D(re+1:re+ne,:);abstract_D(ry+1:ry+ny,:)];  

ddii=pinv(ddi);
K=-ddii*cci;      % [u;w]=cci*x+ddi*f interpreted as f=Kx+ddii*[u;w]
a=abstract_A+abstract_B*K;     % plant coefficients
b=abstract_B*ddii;
c=cco+ddo*K;
d=ddo*ddii;
p=ss(a,b,c,d);
p=minlti(p);
[a,b,c,d]=ssdata(p);
b1=b(:,1:nw);
b2=b(:,nw+1:nw+nu);
c1=c(1:ne,:);
c2=c(ne+1:ne+ny,:);
d22=d(ne+1:ne+ny,nw+1:nw+nu);
d12=d(1:ne,nw+1:nw+nu);
d21=d(ne+1:ne+ny,1:nw);
d11=d(1:ne,1:nw);
d12i=pinv(d12);
d21i=pinv(d21);

if norm(eye(nu)-d12i*d12)>0.1, 
    error('d12 not left invertible'); 
end
if norm(eye(ny)-d21*d21i)>0.1, 
    error('d21 not left invertible'); 
end
if norm(d22)>0.000001*norm(d12), 
	disp('Warning: d22 re-set to zero'); 
end

K0=-d12i*d11*d21i;
if norm(d11+d12*K0*d21)>0.000001*norm(d12)*norm(d21),
	disp('h2hi: d11=d12*K0*d21 not solvable; d11 for h2 modified');
    fprintf('\n');
end
d220=zeros(ny,nu);
d11h2=zeros(ne,nw);
ah2=a+b2*K0*c2;
b1h2=b1+b2*K0*d21;
c1h2=c1+d12*K0*c2;

ph2=ss(ah2,[b1h2 b2],[c1h2;c2],[d11h2 d12;d21 d220]);
[kh2,gh2]=h2syn(ph2,ny,nu);
kh2=kh2+K0;
gh2=gh2+d11+d12*K0*d21;
[ak2,bk2,ck2,dk2]=ssdata(kh2);
gmax=norm(gh2,Inf);

p=ss(a,[b1 b2],[c1;c2],[d11 d12;d21 d220]);
%[khi,ghi]=hinfsyn(p,ny,nu,'GMAX',100,'GMIN',0,'DISPLAY','on');
[khi,ghi]=hinfsyn(p,ny,nu,'DISPLAY','off','METHOD','lmi');










